%{
#include <stdlib.h>


#include "Vector.h"
#include "Plane.h"
#include "Sphere.h"
#include "Triangle.h"
#include "Object.h"
#include "ObjectList.h"
#include "Camera.h"
#include "Color.h"

#include "SphereOpt.h"
#include "ObjectMods.h"
#include "Finish.h"
#include "Pigment.h"
#include "Bitmap.h"
#include "PPMFile.h"

#include "Parameters.h"

/* pigment defaults */
const Color_component PIGMENT_DEFAULT_RED =   127; /* 0.5 of total */
const Color_component PIGMENT_DEFAULT_GREEN = 127;
const Color_component PIGMENT_DEFAULT_BLUE =  127;

/* pole vector defaults */
const double POLE_DEFAULT_X = 0;
const double POLE_DEFAULT_Y = 1;
const double POLE_DEFAULT_Z = 0;

/* equator vector defaults */
const double EQUATOR_DEFAULT_X = 1;
const double EQUATOR_DEFAULT_Y = 0;
const double EQUATOR_DEFAULT_Z = 0;

const double REFLECTION_DEFAULT = 0;
const double DIFFUSE_DEFAULT = 1;

extern char *yytext;
extern int linecount;
extern int yylex(void);

extern Parameters global_parameters;
extern ObjectList global_objectList;
extern Camera     global_camera;
extern Color      global_background;

PPMFile    global_image_file;


int yyerror(char *s);
%}

%union {
    double	value;		        /* for numbers */
    char		*string;	/* for names */
    Vector	Vector;
    Sphere      Sphere;
    Triangle    Triangle;
    Plane       Plane;
    Object      Object;
    ObjectList  ObjectList;
    Camera      Camera;
    Color       Color;
    SphereOpt   SphereOpt;
    ObjectMods  ObjectMods;
    Finish      Finish;
    Pigment     Pigment;
    Bitmap      Bitmap;
    Light       Light;
};

%start scene

%token LBRACE RBRACE LANGLE RANGLE COMMA

%token <value> NUMBER
%token <string> STRING

%token NAME UNCAUGHT

%token KEY_SPHERE KEY_PIGMENT KEY_COLOR
%token KEY_FINISH KEY_TRIANGLE KEY_TRIANGLEPNT
%token KEY_DIFFUSE KEY_REFLECTION
%token KEY_RED KEY_GREEN KEY_BLUE
%token KEY_LOCATION KEY_LOOK KEY_BACKGROUND
%token KEY_CAMERA KEY_RGB KEY_SKY KEY_LIGHT
%token KEY_PLANE KEY_PLANEPNT KEY_IMAGE KEY_PPM
%token KEY_POLE KEY_EQUATOR


%type <ObjectList> scene
%type <Object> item
%type <Plane> plane
%type <Triangle> triangle
%type <Sphere> sphere
%type <Vector> vector
%type <value> number
%type <Camera> camera
%type <Color> color
%type <SphereOpt> sphere_opt
%type <ObjectMods> object_mods  /* opt_pigment, opt_finish*/
%type <Pigment> opt_pigment
%type <Pigment> pigment
%type <Finish> opt_finish       /* opt_diffuse, opt_reflection */
%type <Finish> finish
%type <Bitmap> image
%type <value> reflection
%type <value> diffuse
%type <value> opt_diffuse
%type <value> opt_reflection
%type <Light> light
%type <Color> background

%%

scene	:	
          item { }
        | scene item { }
	;

item:	
          sphere     { global_objectList = 
                           ObjectList_insert( Object_createSphere( $1 ),
                                              global_objectList); } 
	| triangle   { global_objectList  = 
                           ObjectList_insert( Object_createTriangle( $1 ),
                                              global_objectList); }
        | plane      { global_objectList = 
                           ObjectList_insert( Object_createPlane( $1 ),
                                              global_objectList); } 
        | camera     { }
        | background { }
        | light      { global_objectList = 
                           ObjectList_insert( Object_createLight( $1 ),
                                              global_objectList); } 
	;

light:
	KEY_LIGHT LBRACE
	vector
	RBRACE { /*printf("Light at "); Vector_print($3); printf("\n");*/
                 $$ = Light_create( $3 );
	       }
	;

vector:
	LANGLE number COMMA
	number COMMA number RANGLE { /*printf("Vector: %f %f %f\n", $2, $4, $6 );*/
                                     $$ = Vector_createFromCartesian($2, $4, $6); }
	;

plane:
	KEY_PLANE LBRACE
	vector COMMA number
	object_mods
	RBRACE { /*printf("Plane with normal=");
		 Vector_print($3);
		 printf(", dist=%.2f\n", $5);*/
                 $$ = Plane_create($3, Vector_createFromCartesian(0, 0, $5), 
				   ObjectMods_pigment( $6 ));
	       }
;

/* This is not POVray syntax, but is an easier way to specify a plane. */
plane:
	KEY_PLANEPNT LBRACE
	vector COMMA vector
	object_mods
	RBRACE { /*printf("Plane with normal="); Vector_print($3);
		   printf(" at "); Vector_print($5); printf("\n");*/
                 $$ = Plane_create($3, $5, ObjectMods_pigment( $6 ));
	       }
	;

sphere:
	KEY_SPHERE LBRACE vector COMMA number sphere_opt
	object_mods
	RBRACE { /*fprintf(stderr, "Sphere at ");
		   Vector_print($3);
		 printf(", radius %.2f\n", $5);*/
                 $$ = Sphere_create($3, $5, SphereOpt_pole($6), 
                                    SphereOpt_equator($6), 
                                    ObjectMods_pigment( $7 ), ObjectMods_diffuse( $7 ),
                                    ObjectMods_reflection( $7 ));
	       }
	;

sphere_opt:   { $$ = SphereOpt_create(
                     Vector_createFromCartesian(POLE_DEFAULT_X, 
                                                POLE_DEFAULT_Y,
                                                POLE_DEFAULT_Z), 
                     Vector_createFromCartesian(EQUATOR_DEFAULT_X,
                                                EQUATOR_DEFAULT_Y,
                                                EQUATOR_DEFAULT_Z) ); }
	|            
           KEY_POLE vector KEY_EQUATOR vector { $$ = SphereOpt_create( $2, $4 );}
	;

triangle:
	KEY_TRIANGLE LBRACE
	vector COMMA
	vector COMMA
	vector
	object_mods
	RBRACE { /*printf("Triangle\n"); */
                 $$ = Triangle_createFromPointAndVectors($3, $5, $7,
                                                         ObjectMods_pigment($8));
               }
	;

/* Again, this is not POVray, but provides an alternative way of
   specifying a triangle. */
triangle:
	KEY_TRIANGLEPNT LBRACE
	vector COMMA
	vector COMMA
	vector
	object_mods
	RBRACE { /*printf("Triangle(p)\n");*/ 
                 $$ = Triangle_createFromPoints($3, $5, $7, 
                                                ObjectMods_pigment($8));
               }
	;
	
object_mods:
	opt_pigment
        opt_finish { $$ = ObjectMods_create( $1, Finish_diffuse( $2 ), Finish_reflection( $2 ) ); }
	;

opt_pigment:	/* empty */
                { $$ = Pigment_createColor( Color_createFromRGB( PIGMENT_DEFAULT_RED,
                                                                 PIGMENT_DEFAULT_GREEN, 
                                                                 PIGMENT_DEFAULT_BLUE)); }
          
	| pigment { $$ = $1 }
	;

pigment:
	KEY_PIGMENT LBRACE color RBRACE { $$ = Pigment_createColor( $3 );}
      | KEY_PIGMENT LBRACE image RBRACE { $$ = Pigment_createBitmap( $3 ); }
	;

                
image:	KEY_IMAGE LBRACE KEY_PPM STRING RBRACE { global_image_file = 
                                                     PPMFile_openIn( $4 );
                                                 $$ = PPMFile_readBitmap( 
                                                     global_image_file );
                                                 PPMFile_close( 
                                                     global_image_file ); }
	;

opt_finish: { $$ = Finish_create( DIFFUSE_DEFAULT, REFLECTION_DEFAULT ); }
        /* empty */
      | finish { $$ = $1 } 
	;

finish:
	KEY_FINISH LBRACE 
	opt_diffuse
	opt_reflection
	RBRACE { $$ = Finish_create( $3, $4 ); }
	;

opt_diffuse: 	{ $$ = DIFFUSE_DEFAULT }		/* empty */
	| diffuse { $$ = $1 }
	;

diffuse:
	KEY_DIFFUSE number { $$ = $2 }
	;

opt_reflection:	 { $$ = REFLECTION_DEFAULT }		/* empty */
	| reflection { $$ = $1 }
	;

reflection:
	KEY_REFLECTION number { $$ = $2 }
	;

color:
	KEY_COLOR
	KEY_RED number
	KEY_GREEN number
	KEY_BLUE number { $$ = Color_createFromRGB( $3 * COLOR_COMPONENT_MAX, 
                                                    $5 * COLOR_COMPONENT_MAX, 
                                                    $7 * COLOR_COMPONENT_MAX); }
	|
	KEY_COLOR KEY_RGB vector { $$ = Color_createFromRGB( Vector_x( $3 ) * COLOR_COMPONENT_MAX,
                                                             Vector_y( $3 ) * COLOR_COMPONENT_MAX,
                                                             Vector_z( $3 ) * COLOR_COMPONENT_MAX); }
	;

camera:
	KEY_CAMERA LBRACE
	KEY_LOCATION vector
	KEY_SKY vector
	KEY_LOOK vector
	RBRACE { /*printf("Camera\n");*/
                 global_camera = 
                     Camera_create( $4, $8, $6, 
                                    Parameters_zoom( global_parameters ),
                                    Parameters_worldWidth( global_parameters),
                                    Parameters_worldHeight( global_parameters ),
                                    Parameters_pixelWidth( global_parameters ),
                                    Parameters_pixelHeight( global_parameters )
                     ); 
		 /*printf("camera created\n");*/ }
	;

background:
	KEY_BACKGROUND LBRACE
	color
	RBRACE { /*printf("Background\n"); */
                 global_background = $3; }
	;

number:
	NUMBER { $$ = $1; }
;

%%

int
yyerror(char *s) {
  fprintf(stderr, "** line %d:%s\n", linecount, s);
  return 1;
}
