/*************************************************************************\      
|  File: Object.c                                                         |
|                                                                         |
|  Implements the ADT for objects                                         |
|                                                                         |
\*************************************************************************/


#include "Object.h"
#include "Error.h"


/**************************************************************************
  Function:  Object_type
  Arguments: Object
  Result:    ObjectType
  Procedure: returns the type of an object
 *************************************************************************/

static ObjectType Object_type( Object o )
{
    return o.type;
}


/**************************************************************************
  Function:  Object_sphere
  Arguments: Object
  Result:    Sphere
  Procedure: returns a sphere object, fails if the object is not a sphere
 *************************************************************************/

static Sphere Object_sphere( Object o )
{
    if( Object_type( o ) != OBJECT_SPHERE)
        Error_printErrorAndExit( "Object is not a sphere" );

    return o.object.sphere;
}


/**************************************************************************
  Function:  Object_triangle
  Arguments: Object
  Result:    Triangle
  Procedure: returns a triangle object, fails if the object is not a triangle
 *************************************************************************/

static Triangle Object_triangle( Object o )
{
    if( Object_type( o ) != OBJECT_TRIANGLE)
        Error_printErrorAndExit( "Object is not a triangle" );

    return o.object.triangle;
}

/**************************************************************************
  Function:  Object_plane
  Arguments: Object
  Result:    plane
  Procedure: returns a plane object, fails if the object is not a plane
 *************************************************************************/

static Plane Object_plane( Object o )
{
    if( Object_type( o ) != OBJECT_PLANE)
        Error_printErrorAndExit( "Object is not a camera" );

    return o.object.plane;
}


Object Object_createSphere( Sphere s )
{
    Object o;
    
    o.type = OBJECT_SPHERE;
    o.object.sphere = s;
    return o;
}

Object Object_createTriangle( Triangle t )
{
    Object o;

    o.type = OBJECT_TRIANGLE;
    o.object.triangle = t;
    return o;
}

Object Object_createPlane( Plane p )
{
    Object o;

    o.type = OBJECT_PLANE;
    o.object.plane = p;
    return o;
}

bool Object_isRayHitting( Object o , Ray r )
{
    bool isHitting = FALSE; /* to make GCC shut up about non initialized.. */

    switch( Object_type( o ) )
    {
    case OBJECT_SPHERE:
        isHitting = Sphere_isRayHitting( Object_sphere( o ), r );
        break;

    case OBJECT_TRIANGLE:
        isHitting = Triangle_isRayHitting( Object_triangle( o ), r );
        break;

    case OBJECT_PLANE:
        isHitting = Plane_isRayHitting( Object_plane( o ), r );
        break;
    }
    return isHitting;
}

double Object_rayHitPointDistance( Object o, Ray r )
{
    double distance = 0; /* to make GCC shut up about non initialized.. */
    
    switch( Object_type( o ) )
    {
    case OBJECT_SPHERE:
        distance = Sphere_rayHitPointDistance( Object_sphere( o ), r );
        break;

    case OBJECT_TRIANGLE:
        distance = Triangle_rayHitPointDistance( Object_triangle( o ), r );
        break;

    case OBJECT_PLANE:
        distance = Plane_rayHitPointDistance( Object_plane( o ), r );
        break;
    }
    return distance;
}

Color Object_rayColorAtHitPoint( Object o, Ray r )
{
    Color col;
    
    switch( Object_type( o ) )
    {
    case OBJECT_SPHERE:
        col = Sphere_rayColorAtHitPoint( Object_sphere( o ), r );
        break;

    case OBJECT_TRIANGLE:
        col = Triangle_rayColorAtHitPoint( Object_triangle( o ), r );
        break;

    case OBJECT_PLANE:
        col = Plane_rayColorAtHitPoint( Object_plane( o ), r );
        break;
    }
    return col;
}

Vector Object_hitPointNormal( Object o, Ray ray )
{
    Vector vect;

    switch( Object_type( o ) )
    {
    case OBJECT_SPHERE:
	vect = Sphere_hitPointNormal( Object_sphere( o ), ray ); 
	break;
    case OBJECT_TRIANGLE:
	vect = Triangle_hitPointNormal( Object_triangle( o ), ray );
	break;
    case OBJECT_PLANE:
	vect = Plane_hitPointNormal( Object_plane( o ), ray );
	break;
    default:
	break;
    }

    return vect;
}


double Object_diffuse( Object o )
{
    double diffuse = 0;

    switch( Object_type( o ) )
    {
    case OBJECT_SPHERE:
	diffuse = Sphere_diffuse( Object_sphere( o )); 
	break;
    case OBJECT_TRIANGLE:
	diffuse = Triangle_diffuse( Object_triangle( o ));
	break;
    case OBJECT_PLANE:
	diffuse = Plane_diffuse( Object_plane( o ) );
	break;
    default:
	break;
    }

    return diffuse;
}

double Object_reflection( Object o )
{
    double reflection = 0;

    switch( Object_type( o ) )
    {
    case OBJECT_SPHERE:
	reflection = Sphere_reflection( Object_sphere( o )); 
	break;
    case OBJECT_TRIANGLE:
	reflection = Triangle_reflection( Object_triangle( o ));
	break;
    case OBJECT_PLANE:
	reflection = Plane_reflection( Object_plane( o ) );
	break;
    default:
	break;
    }

    return reflection;
}


void Object_print( Object o )
{
    switch( Object_type( o ) )
    {
    case OBJECT_SPHERE:
        Sphere_print( Object_sphere( o ) );
        break;
    case OBJECT_TRIANGLE:
        Triangle_print( Object_triangle( o ) );
        break;
    case OBJECT_PLANE:
        Plane_print( Object_plane( o ) );
        break;
    }
}
