/*************************************************************************\
|  File: Sphere.c                                                         |
|                                                                         |
|  Implemants the sphere ADT                                              | 
|                                                                         |
\*************************************************************************/

#include "Sphere.h"

#include <math.h>

/**************************************************************************
  Function:  Sphere_centre
  Arguments: Sphere
  Result:    Vector
  Procedure: given a sphere return it's centre
 *************************************************************************/

static Vector Sphere_centre( Sphere s )
{
    return s.centre;
}


/**************************************************************************
  Function:  Sphere_pole
  Arguments: Sphere
  Result:    Vector
  Procedure: given a sphere return it's pole vector
 *************************************************************************/

static Vector Sphere_pole( Sphere s )
{
    return s.pole;
}


/**************************************************************************
  Function:  Sphere_centre
  Arguments: Sphere
  Result:    Vector
  Procedure: given a sphere return it's equator vector
 *************************************************************************/

static Vector Sphere_equator( Sphere s )
{
    return s.equator;
}


/**************************************************************************
  Function:  Sphere_radius
  Arguments: Sphere
  Result:    double
  Procedure: given a sphere return it's radius 
 *************************************************************************/

static double Sphere_radius( Sphere s )
{
    return s.radius;
}


/**************************************************************************
  Function:  Sphere_pigment
  Arguments: Sphere
  Result:    Pigment
  Procedure: given a sphere return it's pigment
 *************************************************************************/

Pigment Sphere_pigment( Sphere s )
{
    return s.pigment;
}


Sphere Sphere_create( Vector centre, double radius, Vector pole, 
                      Vector equator, Pigment pigment, 
                      double reflection, double diffuse)
{
    Sphere s = {centre, radius, Vector_normalize(pole), 
                Vector_normalize(equator), pigment,
                reflection, diffuse};

    return s;
}


bool Sphere_isRayHitting( Sphere s, Ray ray )
{
    Vector oc;
    double l2oc;
    double tca;
    double t2hc;
    
    oc = Vector_subtract( Sphere_centre( s ),
                          Ray_origin( ray ));
    l2oc = Vector_dotProduct( oc, oc );
    tca = Vector_dotProduct( oc, Ray_direction( ray ) );
    
    
    t2hc = Sphere_radius( s ) *
           Sphere_radius( s ) -
           l2oc + tca * tca;
    
    /* ray origin is inside sphere => ray hits*/
    if( l2oc < Sphere_radius( s ) * 
        Sphere_radius( s ) )
    {
        return TRUE;
    }
    /* an outside ray is hitting */
    else if( tca + SPHERE_EPS >= 0 && t2hc >= 0 )
    {
        return TRUE;
    }

    return FALSE;
}

double Sphere_rayHitPointDistance( Sphere s , Ray ray )
{
    Vector oc;
    double l2oc;
    double tca;
    double t2hc;
    double distance;
    
    oc = Vector_subtract( Sphere_centre( s ),
                          Ray_origin( ray ) );
    l2oc = Vector_dotProduct( oc, oc );
    tca = Vector_dotProduct( oc, Ray_direction( ray ) );
    
    t2hc = Sphere_radius( s ) *
        Sphere_radius( s ) -
        l2oc + tca * tca;
    
        /* inside hit */
    if( l2oc < Sphere_radius( s ) * 
               Sphere_radius( s ))
    {
        /*fprintf( stderr, "Object %d, a sphere is hit from inside\n", 
          index );*/
        distance = tca + sqrt( t2hc );
    }
    /* outside ray is hitting */
    else
    {
        /*fprintf( stderr, "Object %d, a sphere is hit from outside\n", 
          index );*/
        distance = tca - sqrt( t2hc );
    }

    return distance;  
        
}

Color Sphere_rayColorAtHitPoint( Sphere s, Ray ray )
{
    double phi,temp;
    double theta;
    double u,v; // Coordinates in % of bitmap.
    Bitmap bitmap;
    Color retColor;

    if( Pigment_type( Sphere_pigment( s ) ) == PIGMENT_COLOR )
	retColor = Pigment_color( Sphere_pigment( s ) );
    else // PIGMENT_BITMAP
    {
	phi = acos( -Vector_dotProduct( 
	    Sphere_hitPointNormal( s, ray ),
	    Sphere_pole( s ) ) );
	v = 1 - phi / M_PI;

	if( v == 1.0 || v == 0.0 )
	    u = 0.0;
	else
	{
	    temp = Vector_dotProduct(
		       Sphere_hitPointNormal( s, ray ),
		       Sphere_equator( s ) ) / sin( phi );
	    temp = temp > 1.0 ? 1.0 : temp;
	    temp = temp < -1.0 ? -1.0 : temp;

	    theta = acos( temp ) / (2*M_PI);

	    u = Vector_dotProduct( 
		    Vector_crossProduct( Sphere_equator( s ),
                                         Sphere_pole( s ) ),
		    Sphere_hitPointNormal( s, ray )) > 0 ?
		theta : 1 - theta;
		       
	}

	// Ok, u and v are the % coordinates into the bitmap.
	// Lets return the correct color for the given point.
	bitmap = Pigment_bitmap( Sphere_pigment( s ) );
	retColor = Bitmap_colorAt( bitmap, 
				   (Bitmap_width( bitmap ) - 1) * u,
				   (Bitmap_height( bitmap ) - 1) * v );
	
    }

    return retColor;
}
 

Vector Sphere_hitPointNormal( Sphere s, Ray ray )
{
    Vector oc;
    Vector ri;
    Vector rn;
    double l2oc;
    double tca;
    double t2hc;
    double distance;
    
    oc = Vector_subtract( Sphere_centre( s ),
                          Ray_origin( ray ) );
    l2oc = Vector_dotProduct( oc, oc );
    tca = Vector_dotProduct( oc, Ray_direction( ray ) );
    
    t2hc = Sphere_radius( s ) *
        Sphere_radius( s ) -
        l2oc + tca * tca;

    /* inside hit */
    if( l2oc < Sphere_radius( s ) * Sphere_radius( s ) )
    {
        distance = tca + sqrt( t2hc );

        ri = Vector_add( Ray_origin( ray ),
                     Vector_multiply( distance,
                                      Ray_direction( ray ) ) );

        rn = Vector_multiply( -1.0 / Sphere_radius( s ),
                              Vector_subtract( ri,
                                               Sphere_centre( s ) ) );

    }
    /* outside ray is hitting */
    else
    {
        distance = tca - sqrt( t2hc );

        ri = Vector_add( Ray_origin( ray ),
                     Vector_multiply( distance,
                                      Ray_direction( ray ) ) );

        rn = Vector_multiply( 1.0 / Sphere_radius( s ),
                              Vector_subtract( ri,
                                               Sphere_centre( s ) ) );
        
    }
	
    return rn;
}
           

double Sphere_diffuse( Sphere s )
{
    return s.diffuse;
}

double Sphere_reflection( Sphere s )
{
    return s.reflection;
}

void Sphere_print( Sphere s )
{
    fprintf( stderr, "Sphere at: " );
    Vector_print( Sphere_centre( s ));
    fprintf( stderr, "with radius: %f\n", Sphere_radius( s ));
}




