/*************************************************************************\
|  File: Sphere.h                                                         |
|                                                                         |
|  Defines an abstract datatype for spheres                               |
|  uses stack allocation                                                  |
\*************************************************************************/

#ifndef SPHERE_H
#define SPHERE_H

#include "Vector.h"
#include "Pigment.h"
#include "Ray.h"

#define SPHERE_EPS (double)0.001

/* pole vector defaults */
#define POLE_DEFAULT_X (double)0.0
#define POLE_DEFAULT_Y (double)1.0
#define POLE_DEFAULT_Z (double)0.0

/* equator vector defaults */
#define EQUATOR_DEFAULT_X (double)1.0
#define EQUATOR_DEFAULT_Y (double)0.0
#define EQUATOR_DEFAULT_Z (double)0.0



typedef struct
{
    Vector  centre;
    double  radius;
    Vector  pole;
    Vector  equator;
    Pigment pigment;
    double  reflection;
    double  diffuse;
} Sphere;

/**************************************************************************
  Function:  Sphere_create
  Arguments: vector, double, 2 Vectors, Pigment, 2 doubles
  Result:    Sphere
  Procedure: creates a sphere from centre vector and radius
 *************************************************************************/

Sphere Sphere_create( Vector centre, double radius, Vector pole, Vector equator,
                      Pigment pigment, double reflection, double diffuse);



/**************************************************************************
  Function:  Sphere_pigment
  Arguments: Sphere
  Result:    Pigment
  Procedure: Returns the Pigment of the sphere.
 *************************************************************************/

Pigment Sphere_pigment( Sphere s );

/**************************************************************************
  Function:  Sphere_isRayHitting
  Arguments: Sphere, Ray
  Result:    bool
  Procedure: given a sphere and a ray, returns TRUE if ray is hittnig
 *************************************************************************/

bool Sphere_isRayHitting( Sphere s , Ray ray);


/**************************************************************************
  Function:  Sphere_rayHitPointDistance
  Arguments: Sphere, Ray
  Result:    double
  Procedure: given a sphere and ray, returns hitpoint distance
 *************************************************************************/

double Sphere_rayHitPointDistance( Sphere s, Ray ray );


/**************************************************************************
  Function:  Sphere_rayColorAtHitPoint
  Arguments: Sphere, Ray
  Result:    Color
  Procedure: given a sphere and a ray give color at hitpoint
 *************************************************************************/

Color Sphere_rayColorAtHitPoint( Sphere s, Ray ray );

/**************************************************************************
  Function:  Sphere_hitPointNormal
  Arguments: Sphere, Ray
  Result:    Vector
  Procedure: Given sphere and a ray, calculate the normal.
 *************************************************************************/

Vector Sphere_hitPointNormal( Sphere s, Ray ray );


/**************************************************************************
  Function:  Sphere_diffuse
  Arguments: Sphere
  Result:    double
  Procedure: returns diffuse
 *************************************************************************/

double Sphere_diffuse( Sphere s );


/**************************************************************************
  Function:  Sphere_reflection
  Arguments: Sphere
  Result:    double
  Procedure: returns reflection
 *************************************************************************/

double Sphere_reflection( Sphere s );


/**************************************************************************
  Function:  Sphere_print
  Arguments: Sphere
  Result:    void
  Procedure: prints a sphere
 *************************************************************************/




void Sphere_print( Sphere s );

#endif /* SPHERE_H */

