/*************************************************************************\
|  File: Triangle.c                                                       |
|                                                                         |
|  Implements the triangle ADT                                            | 
|                                                                         |
\*************************************************************************/


#include "Triangle.h"
#include "Plane.h"


/**************************************************************************
  Function:  Triangle_origin
  Arguments: Triangle
  Result:    Vector
  Procedure: returns "origin" of a triangle (corner 0)
 *************************************************************************/

static Vector Triangle_origin( Triangle t )
{
    return t.t0;
}


/**************************************************************************
  Function:  Triangle_vectorA
  Arguments: Triangle
  Result:    Vector
  Procedure: returns "first side" of a triangle
 *************************************************************************/


static Vector Triangle_vectorA( Triangle t )
{
    return t.va;
}


/**************************************************************************
  Function:  Triangle_vectorB
  Arguments: Triangle
  Result:    Vector
  Procedure: returns "second side" of a triangle 
 *************************************************************************/

static Vector Triangle_vectorB( Triangle t )
{
    return t.vb;
}


/**************************************************************************
  Function:  Triangle_pigment
  Arguments: Triangle
  Result:    Pigment
  Procedure: returns pigment of a triangle 
 *************************************************************************/

static Pigment Triangle_pigment( Triangle t )
{
    return t.pigment;
}



Triangle Triangle_createFromPoints( Vector c0, Vector c1, Vector c2, 
                                    Pigment pigment,
                                    double reflection,
                                    double diffuse)
{
    Triangle t = {c0, Vector_subtract(c1, c0), Vector_subtract(c2, c0),
                  pigment, reflection, diffuse};

    return t;
}

Triangle Triangle_createFromPointAndVectors( Vector t0, Vector va, Vector vb,
                                             Pigment pigment,
                                             double reflection,
                                             double diffuse)
{
    Triangle t = {t0, va, vb, pigment, reflection, diffuse};

    return t;
}



bool Triangle_isRayHitting( Triangle tr, Ray ray )
{
    Vector pn;
    Plane  pl;
    Vector q;
    double u;
    double v;
    Vector ri;

    pn = Vector_normalize( Vector_crossProduct( Triangle_vectorA( tr ),
                                                Triangle_vectorB( tr ) ) );
    
    pl = Plane_create( pn , Triangle_origin( tr ), Triangle_pigment( tr ),
                       Triangle_reflection( tr ), Triangle_diffuse( tr ));

    ri = Vector_add( Ray_origin( ray ),
                     Vector_multiply( Plane_rayHitPointDistance( pl, ray ),
                                      Ray_direction( ray ) ) );
 
    if( Plane_isRayHitting( pl, ray ) )
    {
        /*fprintf( stderr, "The plane spanned by a triangle is hit...\n");*/
        q = Vector_subtract( ri, 
                             Triangle_origin( tr ) );
        
        u = ( (Vector_dotProduct( Triangle_vectorB( tr ), 
                                  Triangle_vectorB( tr ) ) *
               Vector_dotProduct( q, 
                                  Triangle_vectorA( tr ) )
              )
              - 
              (Vector_dotProduct( Triangle_vectorA( tr ),
                                  Triangle_vectorB( tr ) ) *
               Vector_dotProduct( q, 
                                  Triangle_vectorB( tr ))
              )
            )
           
            /

            ( (Vector_dotProduct( Triangle_vectorA( tr ),
                                  Triangle_vectorA( tr ) ) *
               Vector_dotProduct( Triangle_vectorB( tr ),
                                  Triangle_vectorB( tr ) ) 
              )
              -
              (Vector_dotProduct( Triangle_vectorA( tr ),
                                  Triangle_vectorB( tr ) ) *
               Vector_dotProduct( Triangle_vectorA( tr ),
                                  Triangle_vectorB( tr ) )
              )
            );
        
        if( 0 < u && u < 1 )
        {
            /*fprintf( stderr, "vector u is OK...");*/
            v = ( Vector_dotProduct( q, 
                                     Triangle_vectorB( tr ) ) 
                  -
                  u * Vector_dotProduct( Triangle_vectorA( tr ),
                                         Triangle_vectorB( tr ) )
                )     
                /
                
                Vector_dotProduct( Triangle_vectorB( tr ),
                                   Triangle_vectorB( tr ) );
             
            /*
            Vector_print( Triangle_vectorA( tr ) );
            fprintf( stderr, " b: ");
            Vector_print( Triangle_vectorB( tr ) );
            fprintf( stderr, "\n" );
            fprintf( stderr, "r0: ");
            Vector_print( Ray_origin( ray ) );
            fprintf( stderr, " rd: ");
            Vector_print( Ray_direction( ray ) );
            fprintf( stderr, "\n");
            fprintf( stderr, "t: %f ", Plane_rayHitPointDistance( pl, ray ) );
            fprintf( stderr, "u: %f v: %f\n", u, v);
            fprintf( stderr, "---\na: ");
            */

            if( 0 < v && v < 1 && u+v <= 1 )
            {
                /*fprintf( stderr, "A triangle is hit...\n" );*/
                return TRUE;
            }
        }
                  
    }
    return FALSE;
}
    
 
double Triangle_rayHitPointDistance( Triangle tr, Ray ray )
{
    Vector pn;

    pn = Vector_normalize( 
        Vector_crossProduct( 
            Triangle_vectorA( tr ),
            Triangle_vectorB( tr ) ) );
     
    return Plane_rayHitPointDistance( Plane_create( pn, 
                                                    Triangle_origin( tr ),
                                                    Triangle_pigment( tr ),
                                                    Triangle_reflection( tr ),
                                                    Triangle_diffuse( tr ) ),
                                      ray );
}


Color Triangle_rayColorAtHitPoint( Triangle tr, Ray ray )
{
    Vector pn;
    Bitmap bitmap;
    double u,v;
    Plane  pl;
    Vector q;
    Vector ri;

    pn = Vector_normalize( Vector_crossProduct( 
        Triangle_vectorA( tr ),
        Triangle_vectorB( tr ) ) );

    if( Pigment_type( Triangle_pigment( tr ) ) == PIGMENT_COLOR )
    {
    
	return Plane_rayColorAtHitPoint( Plane_create( pn,
                                                   Triangle_origin( tr ),
                                                   Triangle_pigment( tr ),
                                                   Triangle_reflection( tr ),
                                                   Triangle_diffuse( tr ) ),
                                     ray );
    }
    else // PIGMENT_BITMAP !
    {
	bitmap = Pigment_bitmap( Triangle_pigment( tr ) );
	// Find the x and y coord into the bitmap

	pn = Vector_normalize( Vector_crossProduct( Triangle_vectorA( tr ),
						    Triangle_vectorB( tr ) ) );
	
	pl = Plane_create( pn , Triangle_origin( tr ), Triangle_pigment( tr ),
			   Triangle_reflection( tr ), Triangle_diffuse( tr ));
	
	ri = Vector_add( Ray_origin( ray ),
			 Vector_multiply( Plane_rayHitPointDistance( pl, ray ),
					  Ray_direction( ray ) ) );
	
	q = Vector_subtract( ri, 
			     Triangle_origin( tr ) );
	
	u = ( (Vector_dotProduct( Triangle_vectorB( tr ), 
				  Triangle_vectorB( tr ) ) *
	       Vector_dotProduct( q, 
				  Triangle_vectorA( tr ) )
	    )
	      - 
	      (Vector_dotProduct( Triangle_vectorA( tr ),
				  Triangle_vectorB( tr ) ) *
	       Vector_dotProduct( q, 
				  Triangle_vectorB( tr ))
		  )
	    )
	    
	    /
	    
	    ( (Vector_dotProduct( Triangle_vectorA( tr ),
				  Triangle_vectorA( tr ) ) *
	       Vector_dotProduct( Triangle_vectorB( tr ),
				  Triangle_vectorB( tr ) ) 
		)
	      -
	      (Vector_dotProduct( Triangle_vectorA( tr ),
				  Triangle_vectorB( tr ) ) *
	       Vector_dotProduct( Triangle_vectorA( tr ),
				  Triangle_vectorB( tr ) )
		  )
		);
	
	v = ( Vector_dotProduct( q, 
				 Triangle_vectorB( tr ) ) 
		      -
	      u * Vector_dotProduct( Triangle_vectorA( tr ),
				     Triangle_vectorB( tr ) )
	    )     
	    /
	    
	    Vector_dotProduct( Triangle_vectorB( tr ),
			       Triangle_vectorB( tr ) );		    
	
	return Bitmap_colorAt( bitmap, 
			       (Bitmap_width( bitmap )-1) * u,
			       (Bitmap_height( bitmap )-1) * v );
    }	
}
    
Vector Triangle_hitPointNormal( Triangle tr, Ray ray )
{
    Vector pn;

    pn = Vector_normalize( Vector_crossProduct( 
        Triangle_vectorA( tr ),
        Triangle_vectorB( tr ) ) );
    
    pn = Plane_hitPointNormal( Plane_create( pn,
					     Triangle_origin( tr ),
					     Triangle_pigment( tr ),
                                             Triangle_reflection( tr ),
                                             Triangle_diffuse( tr )),
			       ray );
    return pn;
}
    
double Triangle_diffuse( Triangle t )
{
    return t.diffuse;
}

double Triangle_reflection( Triangle t )
{
    return t.reflection;
}


void Triangle_print( Triangle t )
{
    fprintf( stderr, "Triangle at: ");
    Vector_print( Triangle_origin( t ) );
    fprintf( stderr, " side A: ");
    Vector_print( Triangle_vectorA( t ) );
    fprintf( stderr, " side B: ");
    Vector_print( Triangle_vectorB( t ) );
    fprintf( stderr, "\n");
}
