/*************************************************************************\
|  File: World.c                                                          |
|                                                                         |
|  Implements the ADT world                                               |
\*************************************************************************/

#include "World.h"
#include "Ray.h"

#include <math.h>


/**************************************************************************
  Function:  World_shade
  Arguments: Vector, Vector, Vector
  Result:    double
  Procedure: given light origin, hit point and hit normal,
             give shade factor
 *************************************************************************/

static double World_shade( Vector l0, Vector ri, Vector rn )
{
    Vector s0;
    Vector sd;
    
    s0 = Vector_add( ri,
                     Vector_multiply( EPS,
                                      rn ) );
    sd = Vector_normalize( Vector_subtract( l0, s0 ) );

    return fabs( Vector_dotProduct( rn, sd ) );
}


/**************************************************************************
  Function:  World_camera
  Arguments: World
  Result:    Camera
  Procedure: returns the camera of a world
 *************************************************************************/

static Camera World_camera( World w )
{
    return w.camera;
}


/**************************************************************************
  Function:  World_objects
  Arguments: World
  Result:    ObjectArray
  Procedure: returns the objects in a world
 *************************************************************************/

static ObjectArray World_objects( World w )
{
    return w.objects;

}


/**************************************************************************
  Function:  World_background
  Arguments: World
  Result:    Color
  Procedure: returns the background colorof a world
 *************************************************************************/

static Color World_background( World w )
{
    return w.background;
}



/**************************************************************************
  Function:  World_lights
  Arguments: World
  Result:    LightArray
  Procedure: returns the ligts in a world
 *************************************************************************/

static LightArray World_lights( World w )
{
    return w.lights;
}



/**************************************************************************
  Function:  World_colorAtHitPoint
  Arguments: World, int, int, int
  Result:    Color
  Procedure: returns the color of the spec pixel given coords and
             reflection depth
 *************************************************************************/

static Color World_colorAtHitPoint( World world, int x, int y, Ray ray,
                                          int reflDepth, 
                                          bool noShadowNoReflection )
{
    Vector      hitPoint;
    Vector      hitNormal;
   
    int         numItems = ObjectArray_numObjects( World_objects( world ) );
    int         numLights = LightArray_numLights( World_lights( world ) );

    int         i,j;
    int         index = 0;
    double      distance = HUGE_VAL;
    double      leastDistance = HUGE_VAL;

    double      shade = 0.0;
    
    Ray         reflRay;
    Ray         lightRay;

    bool        isLit;

    Color       col;

    for( i = 0 ; i < numItems ; i++)
    {
        /*fprintf( stderr, "About to check #%d for hit\n", i );*/
        if( Object_isRayHitting( ObjectArray_objectAt( World_objects( world ), 
                                                       i ), 
                                 ray ) )
        {
            distance = 
                Object_rayHitPointDistance( 
                    ObjectArray_objectAt( World_objects( world ), 
                                          i ), 
                    ray );
            
            if( distance < leastDistance )
            {
                leastDistance = distance;
                index = i;
            }
        }
    }

    if( leastDistance == HUGE_VAL )
    { 
        return World_background( world );
    }
   
  

    /* leastDistance is the distance to the nearest object hit */
    hitPoint = Vector_add( Ray_origin( ray ),
                           Vector_multiply(
                               leastDistance,
                               Ray_direction( ray ) ) );
    
    hitNormal = Object_hitPointNormal( ObjectArray_objectAt(
                                          World_objects( world ), 
                                          index ), 
                                       ray );
    
    if( noShadowNoReflection ) /* Simulate the camera as a light source */
    {  
                /*fprintf( stderr, "%d %f\n", index,
                 World_shade( Ray_origin( ray ), hitPoint, hitNormal ) );*/
        
        return 
            Color_shade(  
                Object_rayColorAtHitPoint( ObjectArray_objectAt( 
                    World_objects( world ),
                    index ),
                                           ray ),
                World_shade( Ray_origin( ray ), hitPoint, hitNormal ) );
            
    }
    else
    {
        
        /* iterate over light sources */
        for( i = 0 ; i < numLights ; i++ )
        {
            /* create a ray from the hit point to a light */
            lightRay = Ray_create( Vector_add( 
                                       hitPoint,
                                       Vector_multiply(
                                           EPS,
                                           Vector_normalize( hitNormal ) ) ),
                                   Vector_subtract(
                                       Light_position(LightArray_lightAt(
                                           World_lights( world ), i ) ),
                                       Vector_add( 
                                       hitPoint,
                                       Vector_multiply(
                                           EPS,
                                           Vector_normalize( hitNormal ) ) ) ) );
            
            isLit = TRUE;
            j = 0;
            while( ( j < numItems ) && isLit )
            {
                if(Object_isRayHitting( ObjectArray_objectAt(
                    World_objects( world ), j ),
                                        lightRay ) )
                {
                    distance = Object_rayHitPointDistance(
                                   ObjectArray_objectAt(
                                       World_objects( world ), j ),
                                   lightRay );
                    /* check to see if object lies between hit point
                       and light source, object is shadowin the light */
                    if( distance < Vector_length( 
                                     Vector_subtract( 
                                         Light_position(LightArray_lightAt(
                                             World_lights( world ), i ) ),
                                         hitPoint ) ) )
                    {
                        isLit = FALSE;
                    }
                }
                j++;
            }


            /*if(!isLit) fprintf( stderr, "dark\n" );
            else fprintf( stderr, "light on me\n" );*/
            
            if( isLit )
            {
                /*fprintf( stderr, "a" );*/
                shade += 
                    World_shade( 
                        Light_position( 
                            LightArray_lightAt(
                                World_lights( world ), i ) ),
                        hitPoint, 
                        hitNormal );
            }
            
        }
        
        
        /*if( shade > 0.0 )
        {
            fprintf( stderr, "Pixel: %d %d is lit at refl.depth %d\n",
                     x,y,reflDepth );
        }*/
        /*fprintf( stderr, "shade: %f\n", shade );*/

        
        /* Base case: don't take reflection into acount, 
           but shade the accumulated light from light sources */
        if( reflDepth == 0 )
        {
            return
                Color_shade( Object_rayColorAtHitPoint( 
                                    ObjectArray_objectAt(
                                        World_objects( world ), index ), 
                                    ray ),
                                shade *
                                Object_diffuse(
                                    ObjectArray_objectAt(
                                        World_objects( world ), index ) ) );
        }
        else
        {

            /* calculate reflected ray */
            /* move EPS in the normal direction, to avoid rounding errors */
            reflRay = Ray_create( Vector_add( hitPoint,
                                              Vector_multiply( EPS,
                                                               hitNormal ) ),
                                  Vector_subtract(
                                      Ray_direction( ray ),
                                      Vector_multiply( 
                                          2 *
                                          Vector_dotProduct(
                                              hitNormal,
                                              Ray_direction( ray ) ) ,
                                          hitNormal ) ) );

            reflDepth--;
            col = World_colorAtHitPoint(
                world, x, y, reflRay, reflDepth, noShadowNoReflection );

            
                                                     
            return
                Color_combine( 
                    Color_shade(
                        col,
                        Object_reflection(
                            ObjectArray_objectAt(
                                World_objects( world ), 
                                index ) ) ),
                    
                    Color_shade( 
                        Object_rayColorAtHitPoint( 
                            ObjectArray_objectAt(
                                World_objects( world ), index ), ray ),
                        
                        shade *
                        Object_diffuse(
                            ObjectArray_objectAt(
                                World_objects( world ), 
                                index ) ) ) );
               
            
        }
       
    }
}





World World_create( ObjectList objects, LightList lights, 
                    Camera cam, Color background )
{
    World world;
       
    world.camera =     cam;
    world.lights =     LightArray_createFromList( lights );
    world.objects =    ObjectArray_createFromList( objects );
    world.background = background;

    return world;
}

void World_free( World world )
{
    LightArray_destroy( world.lights );
    ObjectArray_destroy( world.objects );
}


Color World_colorOfPixel( World world, int x, int y, int reflDepth,
                          bool noShadowNoReflection )
{
    Ray ray;
    Vector c, cRight, cDown, cCorner;
    Vector rayDirection;
    
    c = Vector_normalize( Vector_subtract(
                                 Camera_lookAt( World_camera( world ) ),
                                 Camera_location( World_camera( world ) ) ) );
                          
    cRight = Vector_normalize( Vector_crossProduct( 
        Camera_upVector( World_camera( world ) ), c ) );
    cDown = Vector_normalize( Vector_crossProduct( cRight, c ) );
    cCorner = Vector_add( 
        Camera_location( World_camera( world ) ), 
        Vector_subtract( 
            Vector_multiply( 
                Camera_zoom( World_camera( world ) ), 
                c ),
            Vector_add( 
                Vector_multiply( Camera_worldWidth( 
                                    World_camera( world ) ) / 2, 
                                 cRight ),
                Vector_multiply( Camera_worldHeight( 
                                    World_camera( world ) ) / 2,  
                                 cDown ) ) ) );
    rayDirection = 
        Vector_normalize( 
            Vector_subtract( 
                Vector_add( 
                    Vector_add( 
                        Vector_multiply( ( Camera_worldWidth( 
                                                    World_camera( world ) ) /  
                                           Camera_pixelWidth( 
                                               World_camera( world ) ) ) 
                                         * x, 
                                         cRight ),
                        Vector_multiply( ( Camera_worldHeight( 
                                                  World_camera( world ) ) / 
                                           Camera_pixelHeight( 
                                               World_camera( world ) ) ) * y, 
                                         cDown ) ),
                    cCorner ), 
                Camera_location( World_camera( world ) ) ) );
    
    ray = Ray_create( Camera_location( World_camera( world ) ), rayDirection );

    /*fprintf( stderr, "%d %d\n", x,y );*/

    return World_colorAtHitPoint( world, x, y, ray, reflDepth,
                                                     noShadowNoReflection );
}





void World_print( World world )
{
    fprintf( stderr, "\n\n World: \n\n " );
    ObjectArray_print( world.objects );
    LightArray_print( world.lights );
    Camera_print( World_camera( world ) );
}

